/* $Header: /usr2/brugere/jbln/uc133/src/RCS/uc.c,v 1.6 1995/03/06 23:18:47 jbln Exp jbln $ */

#include <malloc.h>
#include <string.h>
#include <ctype.h>
#include <dirent.h>
#include <pwd.h>
#include <grp.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#include "uc.h"
#include "keys.h"
#include "hyper.h"
#include "cfgfile.h"
#include "input.h"


/************************************************************
   DEFINITIONS
************************************************************/

/* ===== FILEMODE ======================================== */
int fmode_values[10] = {0400,0200,0100,040,020,010,04,02,01};
char fmode_valnames[10] = "rwxrwxrwx";
char *ftype_names[4] = {"Directory", "file", "device", ""};

#define Idirmode (S_IEXEC | S_IREAD | S_IWRITE | S_IRGRP | \
		  S_IROTH | S_IWGRP | S_IXGRP | S_IXOTH)


/* ===== DEFAULT SETUP =================================== */
FILESPEC uc[2] 
   = { {".", ".", (FILEINFO *)NULL, 0, 0, 0,-1,0,0, 1,0, 16, SORT_NAME, 0,
	0, Idirmode, DEF_SHOWHIDDEN, 0}, 
       {".", ".", (FILEINFO *)NULL, 0, 0, 0,-1,0,0, 1,0, 16, SORT_NAME, 0,
	0, Idirmode, DEF_SHOWHIDDEN, 0} };


/* ===== HELP AND INPUT ================================== */
char *buttons[10] 
   = { "Chmod","Chown","View","Edit","Copy",
       "RenMov","Mkdir","Delete","Home","Quit" };

static char input_str[201] = {""};
static INP *lineinp = NULL;


/* ===== COMMAND HISTORY ================================= */
char command_stack[HISTORY][201];
int cstack_front=0;
int cstack_back=0;
int cstack_cur=0;


/* ===== DATE CONVERSIONS ================================ */
int thisyear;
char *monname[12] =
   {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

char *subst_tilde();

/* ===== ALIASES ========================================= */
ASSOC *diralias = NULL;
int daliasnum = 0;


/* ===== TERMINAL INFORMATION ============================ */
TERMINFO terminal =
   { KEY_UP, KEY_DOWN, KEY_LEFT, KEY_RIGHT, key_enter,
     KEY_PPAGE, KEY_NPAGE, 8 };

char *termnames[TERMKEYNUM] =
   { "up", "down", "left", "right", "enter", "pageup", "pagedown",
     "backspace" };

char *configvals[CONFIGNUM];


/* ===== SCREEN SETUP ==================================== */
char *screennames[SCREENNUM] =
   { "showhidden", "sorting", "showmode", "size"};


/* ===== PREFERENCES  ==================================== */
char *prefnames[PREFNUM] =
   { "editor", "viewer", "savehistory", "shell" };

char *prefvals[PREFNUM];


/* ===== USERKEYS ======================================== */
char *userkeys[UKNUM] =
   {NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   };

/* ===== OPTIONS FOR DEFULAT SHELL ======================= */
static char *prefshellopt = {""};


/* ===== OTHER =========================================== */
static char ucgrpmask[80] = {"*"};


/************************************************************
   MAIN & SETUP
************************************************************/

main(ac,av)
int ac;
char **av;
{
   int c; 

   setup();
   test_args(ac, av);
   run_uc();
   move(23,0);
   refresh();

   close_curses();
   config_close();
   ff_clear();
   writetree();

   printf("\n\n");
   return(0);
}


test_args(ac, av)
int ac;
char **av;
{
   int n;

   for (n=1 ; n<ac ; n++)
   {
      if (strcmp(av[n], "key") == 0)
	 keyboard_setup();
   }
}


setup()
{
   time_t now;
   struct tm *tt;
   int n;

   time(&now);
   tt = localtime(&now);
   thisyear = tt->tm_year;

   config_setup();
   find_cmd_options();
   initscr();
   setup_curses();
   if ( (lineinp=inp_new(1,22,input_str,50,200,0)) == (INP *)NULL)
   {
      close_curses();
      printf("Could not create input\n");
      exit(1);
   }
}


find_cmd_options()
{
   char *p;

   p = prefvals[PREFSHELL];

      /* Find first nonblank */
   while (!isspace(*p)  &&  *p!=0)
      p++;
   if (*p == 0)
      return(0);

      /* Remove options from shellname */
   *p = 0;
   p++;

      /* Remove blanks between */
   while (isspace(*p))
      p++;

      /* Rest is option string */
   prefshellopt = p;

   return(1);
}


/************************************************************
   PROGRAM
************************************************************/

run_uc()
{
   int ok=1;
   int c;
   int pos=0;
   int y=0, oldy=0;
   char *p;

   clear();
   draw_screen(0);
   draw_screen(1);
   draw_buttons();
   test_maketree(0,0);
   test_maketree(1,0);

   update_bscreens();

   cursor(pos);
   do
   {
      if (uc[pos].oldpos != uc[pos].pos  ||
	  uc[pos].oldoffset != uc[pos].offset)
      {
	 if (uc[pos].oldoffset != uc[pos].offset)
	    print_files(pos);
         else
	 if (uc[pos].oldpos != -1)
	    print_current(pos,uc[pos].oldpos,0);
	 uc[pos].oldpos = uc[pos].pos;
	 uc[pos].oldoffset = uc[pos].offset;
	 print_current(pos,uc[pos].pos,1);

         if (uc[pos].showmode == SHOWTREE)
         {
            chdir(GETINFO(pos).tree.path);
            getcwd(uc[pos].path, UCPATHSIZE);
         }
      }

      path_prompt(pos);

      c = ti_getch();
      if (c>=32  &&  c<127  ||  c==terminal.backspace  ||
	  c==terminal.left  ||  c==terminal.right      ||
	  c==key_cc)
      {
	 if (inp_addchr(lineinp, c) == INPESC)
	    inp_update(lineinp);
      }
      else
      if (c == terminal.pageup  ||  c == key_cb)
         y = backward(pos,y);
      else
      if (c == terminal.pagedown  ||  c == key_cf)
         y = forward(pos,y);
      else
      if (c == terminal.up)
         y--;
      else
      if (c == terminal.down)
      {
         if (y+uc[pos].offset+1<uc[pos].filenum)
	    y++;
      }
      else
      if (c == terminal.enter)
         y = do_enter(pos,y);
      else
      switch(c)
      {
	 case key_tab  :  print_current(pos,y,0);
			  if (chdir(uc[N(pos)].path)==0)
			  {
			     pos = N(pos);
			     y = uc[pos].pos;
                          }
			  else
                          {
			     print_error();
			     pos = N(pos);
			     strcpy(uc[pos].path, (char *)getenv("HOME"));
                             shortpath(pos);
			     uc[pos].pos = 0;
			     uc[pos].offset = 0;
			     refresh();
			     update_screen(pos);
			     refresh();
			     y = uc[pos].pos;
			  }
			  print_current(pos,y,1);
			  break;
	 case key_cc   :  inp_clear(lineinp);
			  break;
	 case key_cd   :  show_fileinfo(pos);
			  cursor(pos);
			  break;
	 case key_ce   :  do_history_back();
			  break;
         case key_cg   :  mark_one(pos);
			  y = uc[pos].pos;
			  break;
         case key_cj   :  inp_addstr(lineinp, GETINFO(pos).name);
			  break;
	 case key_ck   :  inp_addstr(lineinp, uc[pos].path);
			  break;
	 case key_cl   :  y = to_last(pos,y);
			  break;
	 case key_cn   :  y = change_dir(pos);
			  cursor(pos);
			  break;
	 case key_co   :  y = search_file(pos, y);
			  break;
	 case key_cp   :  test_userkeys(pos);
			  break;
         case key_cr   :  test_maketree(pos,1);
         		  update_screen(pos);
			  cursor(pos);
			  if (strcmp(uc[0].path, uc[1].path)==0)
			     update_screen(N(pos));
			  y = uc[pos].pos;
			  break;
	 case key_ct   :  y = to_top(pos,y);
			  break;
	 case key_cu   :  swappanels(pos);
			  pos = N(pos);
			  y = uc[pos].pos;
			  break;
	 case key_cx   :  do_history_forw();
			  break;
	 case key_esc  :
	 case KEY_F(1) :  ok = do_functionkeys(pos);
			  y = uc[pos].pos;
			  break;
	 case KEY_F(2) :  do_chown(pos);
			  y = uc[pos].pos;
			  break;
	 case KEY_F(3) :  do_view(pos);
			  y = uc[pos].pos;
			  break;
	 case KEY_F(4) :  do_edit(pos);
			  y = uc[pos].pos;
			  break;
	 case KEY_F(5) :  do_copy(pos);
			  y = uc[pos].pos;
			  break;
	 case KEY_F(6) :  do_move(pos);
			  y = uc[pos].pos;
			  break;
	 case KEY_F(7) :  do_mkdir(pos);
			  y = uc[pos].pos;
			  break;
	 case KEY_F(8) :  do_delete(pos);
			  y = uc[pos].pos;
			  break;
	 case KEY_F(9) :  do_tohomedir(pos);
			  y = uc[pos].pos;
			  break;
	 case KEY_F(10):  ok = 0;
			  break;
      }
      if (y<0)
      {
	 if (uc[pos].offset>0)
	    uc[pos].offset--;
         y=0;
      }
      if (y>=uc[pos].size)
      {
	 if (uc[pos].offset+y < uc[pos].filenum)
	    uc[pos].offset++;
         y=uc[pos].size-1;
      }
      uc[pos].pos = y;
   }
   while (ok);
}


path_prompt(pos)
int pos;
{      
   char *path;
   int len;

   if (uc[pos].showmode == SHOWTREE)
   {
      path = uc[pos].files[uc[pos].pos+uc[pos].offset].tree.path;
      if ((len=strlen(path)) > 70)
         path = &path[len-70];
   }
   else
      path = uc[pos].spath;

   len = strlen(path)+1;
   inp_repos(lineinp, len, 22, 75-len);
   prompt(path, ">");
   inp_update(lineinp);
   refresh();
}


forward(pos, y)
int pos,y;
{
   uc[pos].offset += uc[pos].size;
   if (uc[pos].offset >= uc[pos].filenum)
   {
      uc[pos].offset = uc[pos].filenum - uc[pos].size;
      if (uc[pos].offset < 0)
	 uc[pos].offset = 0;
      y = uc[pos].filenum - uc[pos].offset - 1;
      if (y >= uc[pos].size)
	 y = uc[pos].size - 1;
   }
   else
   {
      y = 0;
   }
   print_files(pos);
   uc[pos].pos = y;
   uc[pos].oldpos = y;
   uc[pos].oldoffset = uc[pos].offset;
   cursor(pos);
   return(y);
}


backward(pos, y)
int pos,y;
{
   if (uc[pos].offset >= uc[pos].size)
   {
      uc[pos].offset -= uc[pos].size;
      y = uc[pos].filenum - uc[pos].offset - 1;
      if (y >= uc[pos].size)
	 y = uc[pos].size - 1;
   }
   else
   {
      if (uc[pos].offset + y < uc[pos].size)
	 y=0;
      uc[pos].offset = 0;
      if (y!=0)
      {
         y = uc[pos].filenum - uc[pos].offset - 1;
         if (y >= uc[pos].size)
	    y = uc[pos].size - 1;
      }
   }
   print_files(pos);
   uc[pos].pos = y;
   uc[pos].oldoffset = uc[pos].offset;
   cursor(pos);
   return(y);
}


to_top(pos, y)
int pos, y;
{
   uc[pos].pos = 0;
   uc[pos].offset = 0;
   return(0);
}


to_last(pos, y)
int pos, y;
{
   if (uc[pos].filenum <= uc[pos].size)
   {
      uc[pos].offset = 0;
      uc[pos].pos = uc[pos].filenum-1;
   }
   else
   {
      uc[pos].offset = uc[pos].filenum - uc[pos].size;
      uc[pos].pos = uc[pos].size-1;
   }

   return(uc[pos].pos);
}



/************************************************************
   DRAW
************************************************************/

clr_display()
{
   int n;
   for (n=1 ; n<23 ; n++)
   {
      move(n,0);
      clrtoeol();
   }
}


update_bscreens()
{
   draw_buttons();
   update_screen(0);
   update_screen(1);
   refresh();
}


update2_bscreens()
{
   draw_screen(0);
   print_files(0);
   print_dir(0);
   draw_screen(1);
   print_files(1);
   print_dir(1);
   refresh();
}


update_screen(pos)
int pos;
{
   draw_screen(pos);
   setup_ucfiles(pos);
   mkt_findpos(pos);
   print_files(pos);
   print_dir(pos);
}


draw_screen(pos)
int pos;
{
   int p,n;

   if (uc[pos].showmode == SHOWWIDE)
   {
      draw_screen_wide(pos);
      return(1);
   }

   p = pos*40;
   move(1,p);
   printw("+======================================+");
   move(21,p);
   printw("+======================================+");
   move(19,p+1);
   printw("--------------------------------------");
   for (n=2 ; n<=20 ; n++)
   {
      move(n,p);
      printw("I ");
      move(n,p+38);
      printw(" I");
   }
   attrset(A_BOLD);
   move(2,p+2);
   printw("Name");
   move(2,p+21);
   printw(" Size");
   move(2,p+30);
   printw("  Mode");
   attroff(A_BOLD);

   return(1);
}


draw_screen_wide(pos)
int pos;
{
   int n;

   move(1+pos*11,0);
   printw("+==============================================================================+");
   move(11+pos*10,0);
   printw("+==============================================================================+");

   for (n=0+pos ; n<=8 ; n++)
   {
      move(n+pos*10+2,0);
      printw("I");
      move(n+pos*10+2,79);
      printw("I");
   }

   attrset(A_BOLD);
   move(2,2);
   printw("Name");
   move(2,21);
   printw(" Size");
   move(2,30);
   printw("  Mode");
   move(2,40);
   printw("  Owner");
   move(2,50);
   printw("  Group");
   move(2,60);
   printw(" Date");
   move(2,70);
   printw(" Time");
   attroff(A_BOLD);
}


draw_buttons()
{
   int n;
   for (n=0 ; n<10 ; n++)
   {
      move(23,n*8);
      printw("%d", (n+1)%10);
      attrset(A_REVERSE);
      printw("%-6s", buttons[n]);
      attroff(A_REVERSE);
   }
   attrset(A_REVERSE);
   move(0,0);
   printw("F1+(L/C/O/H/R) =>    Left    Commands    Options    Help    Right               ");
   attroff(A_REVERSE);
}


print_files(pos)
int pos;
{
   int n;

   for (n=0 ; n<uc[pos].size ; n++)
   {
      if (uc[pos].showmode==SHOWNORM  ||  uc[pos].showmode==SHOWTREE)
      {
         move(n+3,pos*40+2);
         printw("                                     ");
         if (n+uc[pos].offset < uc[pos].filenum)
            print_one_file(n+3, pos, uc[pos].files[uc[pos].offset+n],0);
      }
      else
      if (uc[pos].showmode == SHOWWIDE)
      {
         move(n+3+pos*10,2);
         printw("                                                                          ");
         if (n+uc[pos].offset < uc[pos].filenum)
            print_one_file(n+3, pos, uc[pos].files[uc[pos].offset+n],0);
      }
   }
}


print_one_file(y,pos, info, do_rev)
int y, pos;
FILEINFO info;
int do_rev;
{
   char modestr[11];
   int len;

   if (uc[pos].showmode == SHOWTREE)
   {
      mkt_show_one(y,pos,&info,do_rev);
      return(1);
   }

   if (do_rev)
      attrset(A_REVERSE);

   if (pos==1  &&  uc[pos].showmode == SHOWWIDE)
   {
      y += 10;
      pos = 0;
   }

   move(y, pos*40+2);
   if (info.mark)
      attron(A_BOLD);
   printw("%-16.16s", info.name);
   if (info.islink)
   {
      if ((len=strlen(info.name)) < 16)
         mvaddch(y,pos*40+len+2,'@');
      else
         mvaddch(y,pos*40+18,'@');
   }
   if (info.mark)
      attroff(A_BOLD);

   build_modestr(modestr, info.mode);
   move(y, pos*40+29);
   printw("%s", modestr);
   move(y, pos*40+20);

   if (info.type == FTYPE_NORM)
      printw("%8ld", info.size);
   else
   if (info.type == FTYPE_FIFO)
      printw("> FIFO <");
   else
   if (info.type == FTYPE_DIR)
   {
      if (strcmp(info.name, "..")==0)
         printw(">UP-DIR<");
      else
	 printw(">SUBDIR<");
   }
   else
   if (info.type == FTYPE_DEV)
      printw(">DEVICE<");
   else
      printw(">OTHER <");
  
   if (uc[pos].showmode == SHOWWIDE)
   {
      mvprintw(y, 40, "%8s", info.usrname);
      mvprintw(y, 50, "%8s", info.grpname);
      mvprintw(y, 60, "%3s %02d", monname[info.fdate.tm_mon], info.fdate.tm_mday);
      if (info.fdate.tm_year != thisyear)
      {
	 mvprintw(y, 70, " %4d", info.fdate.tm_year+1900);
      }
      else
      {
	 mvprintw(y, 70, "%02d:%02d", info.fdate.tm_hour, info.fdate.tm_min);
      }
   }

   if (do_rev)
      attroff(A_REVERSE);

   return(1);
}


print_grpsize(pos)
int pos;
{
   char valstr[25];

   if (uc[pos].showmode != SHOWWIDE)
      mvprintw(20,pos*40+2, "                              ");
   else
      mvprintw(11+pos*10,2, "==============================");

   if (uc[pos].grpcou > 0)
   {
      valstr[0] = 0;
      val2str(uc[pos].grpsize, valstr);
      if (uc[pos].showmode == SHOWNORM)
      {
         move(20, pos*40+2);
         printw("%s bytes in %d files", valstr, uc[pos].grpcou);
      }
      else
      {
	 move(11+pos*10, 2);
         printw(" %s bytes in %d files ======", valstr, uc[pos].grpcou);
      }
   }
}


val2str(val, str)
long val;
char *str;
{
   long nval;
   char tmp[10];

   nval = val/1000;

   if (nval == 0)
   {
      sprintf(tmp, "%d", val%1000);
      strcat(str, tmp);
      return(0);
   }

   sprintf(tmp, "%03d", val%1000);
   val2str(nval, str);
   strcat(str, ".");
   strcat(str, tmp);

   return(1);
}


prompt(str1, str2)
char *str1, *str2;
{
   move(22,0);
   clrtoeol();
   printw("%s%s",str1, str2);
   refresh();
}


print_current(pos, y, stat)
int pos, y, stat;
{
   print_one_file(y+3, pos, uc[pos].files[y+uc[pos].offset], stat);
}


cursor(pos)
int pos;
{
   print_one_file(uc[pos].pos+3, pos, GETINFO(pos), 1);
}


print_error()
{
   move(22,0);
   clrtoeol();
   if (errno >= sys_nerr)
      printw("Unknown error : code %d", errno);
   else
      printw("ERROR : %s.", sys_errlist[errno]);
   refresh();
   ti_getch();
}


print_dir(pos)
int pos;
{
   char str[60];
   char *dir;
   int p,n,len,delsize;
   int lastsl=15,newsl=15;

   if ( (len=strlen(uc[pos].path)) <= 34)
      strcpy(str, uc[pos].path);
   else
   {
      dir = uc[pos].path;
      p=0;
      while (p<15)
      {
	 if (dir[p]=='/')
	    lastsl = p;
	 p++;
      }
      p = len-1;
      newsl = len - 15;
      if (newsl < 0)
	 newsl = len-1;
      while (len-p < 15)
      {
	 if (dir[p]=='/')
	    newsl = p;
         p--;
      }
      for (p=0,n=0 ; p<=lastsl ; p++,n++)
	 str[n]=dir[p];
      for (p=0 ; p<4 ; p++,n++)
         str[n]='.';
      for (p=newsl ; p<len ; p++, n++)
	 str[n]=dir[p];
      str[n]=0;
   }
   len = strlen(str);
   delsize = (34-len)/2;
   if (uc[pos].showmode != SHOWWIDE)
      move(1, pos*40+2);
   else
      move(1+pos*11, 2);
   printw("%*.*s %s %*.*s", delsize,delsize,"========================",
			    str,
			    delsize,delsize,"========================");
}


/************************************************************
   FINDING FILES
************************************************************/

find_files(pos)
int pos;
{
   if (uc[pos].showmode != SHOWTREE)
      find_files_norm(pos);
}


find_files_norm(pos)
int pos;
{
   DIR *dirp;
   int n, fnum=0;
   int type;
   int use_this;
   char tmp[200], dummy[5];
   struct dirent *dp;
   struct stat fstat;
   struct group *grpe;
   struct passwd *pwde;

   reset_ucfiles(pos);

   dirp = opendir(uc[pos].path);
   if (dirp != (DIR *)NULL)
   {
      while ( (dp=readdir(dirp)) != (struct dirent *)NULL )
      {
	 use_this = 1;

	 if (dp->d_name[0] == '.'  &&  dp->d_name[1] != '.'  &&
	     !uc[pos].showhidden)
	       use_this = 0;

	 if (dp->d_name[0] == '.'  &&  dp->d_name[1] == 0)
	    use_this = 0;

         if (use_this)
         {
	    if (uc[pos].allocnum <= uc[pos].filenum)
	    {
	       uc[pos].allocnum += FALLOCSIZE;
	       if (uc[pos].files == (FILEINFO *)NULL)
		  uc[pos].files = (FILEINFO *)malloc(sizeof(FILEINFO) * uc[pos].allocnum);
               else
		  uc[pos].files = (FILEINFO *)realloc(uc[pos].files, sizeof(FILEINFO) * uc[pos].allocnum);
	    }

	    uc[pos].files[uc[pos].filenum].name = strdup(dp->d_name);
            sprintf(tmp,"%s/%s", uc[pos].path, dp->d_name);
            stat(tmp, &fstat);
            uc[pos].files[uc[pos].filenum].size = fstat.st_size;
            uc[pos].files[uc[pos].filenum].mode = fstat.st_mode;
            uc[pos].files[uc[pos].filenum].linknum = fstat.st_nlink;
            uc[pos].files[uc[pos].filenum].uid = fstat.st_uid;
            uc[pos].files[uc[pos].filenum].gid = fstat.st_gid;
            uc[pos].files[uc[pos].filenum].mark = 0;
            uc[pos].files[uc[pos].filenum].tree.descr = NULL;
            uc[pos].files[uc[pos].filenum].tree.path = NULL;

            if (uc[pos].showmode == SHOWWIDE)
	    {
               if ( (grpe=getgrgid(fstat.st_gid)) != NULL )
	          strcpy(uc[pos].files[uc[pos].filenum].grpname, grpe->gr_name);
               else
                  strcpy(uc[pos].files[uc[pos].filenum].grpname, "----"); 

               if ( (pwde=getpwuid(fstat.st_uid)) != NULL )
	          strcpy(uc[pos].files[uc[pos].filenum].usrname, pwde->pw_name);
               else
                  strcpy(uc[pos].files[uc[pos].filenum].usrname, "----"); 
  	    }
	    else
	    {
               strcpy(uc[pos].files[uc[pos].filenum].grpname, "----"); 
               strcpy(uc[pos].files[uc[pos].filenum].usrname, "----"); 
	    }

            uc[pos].files[uc[pos].filenum].filetime = fstat.st_mtime;
            uc[pos].files[uc[pos].filenum].fdate = *localtime(&fstat.st_mtime);

            if ((fstat.st_mode & FMODE_NORM) == FMODE_NORM)
	       type = FTYPE_NORM;
            else
            if ((fstat.st_mode & S_IFIFO) == S_IFIFO)
            {
               type = FTYPE_FIFO;
            }
            else
            if ( (fstat.st_mode & FMODE_SPEC) == FMODE_DIR)
	    {
	       type = FTYPE_DIR;
               uc[pos].files[uc[pos].filenum].size = 0;
            }
            else
            if ( (fstat.st_mode & FMODE_SPEC) == FMODE_CHAR  ||  
	         (fstat.st_mode & FMODE_SPEC) == FMODE_BLCK)
	       type = FTYPE_DEV;
            else
	       type = FTYPE_OTHER;

#ifdef NOREADLINK
            uc[pos].files[uc[pos].filenum].islink = 0;
#else            
            if (readlink(tmp, dummy, 5) >= 0)
               uc[pos].files[uc[pos].filenum].islink = 1;
            else
               uc[pos].files[uc[pos].filenum].islink = 0;
#endif /* NOREADLINK */

            uc[pos].files[uc[pos].filenum].type = type;
            uc[pos].files[uc[pos].filenum].pos = fnum++;
            uc[pos].filenum++;
         }
      }
      closedir(dirp);
   }
   else
   {
      if (uc[pos].allocnum == 0)
      {
         uc[pos].files = (FILEINFO *)malloc(sizeof(FILEINFO)*FALLOCSIZE);
         uc[pos].allocnum = FALLOCSIZE;
      }
      uc[pos].files[0].name = strdup("..");
      stat("..", &fstat);
      uc[pos].files[0].type = FTYPE_DIR;
      uc[pos].files[0].size = fstat.st_size;
      uc[pos].files[0].mode = fstat.st_mode;
      uc[pos].files[0].linknum = fstat.st_nlink;
      uc[pos].files[0].uid = fstat.st_uid;
      uc[pos].files[0].gid = fstat.st_gid;
      uc[pos].files[0].mark = 0;
      uc[pos].files[0].tree.descr = NULL;
      uc[pos].files[0].tree.path = NULL;
      uc[pos].filenum = 1;
   }

   sort_files(pos);

   if (uc[pos].offset > uc[pos].filenum)
      uc[pos].offset = uc[pos].filenum-1;
   if (uc[pos].pos+uc[pos].offset >= uc[pos].filenum)
      uc[pos].pos = uc[pos].filenum - uc[pos].offset - 1;
   if (uc[pos].pos < 0)
      uc[pos].pos = 0;
   if (uc[pos].pos >= uc[pos].size)
      uc[pos].pos = uc[pos].size-1;
   uc[pos].oldpos = uc[pos].pos;
}


reset_ucfiles(pos)
int pos;
{
   int n;

   if (uc[pos].files != (FILEINFO *)NULL)
   {
      for (n=0 ; n<uc[pos].filenum ; n++)
      {
         if (uc[pos].files[n].name != NULL)
            free(uc[pos].files[n].name);
         if (uc[pos].files[n].tree.descr != NULL)
            free(uc[pos].files[n].tree.descr);
         if (uc[pos].files[n].tree.path != NULL)
            free(uc[pos].files[n].tree.path);
      }
   }

   uc[pos].filenum = 0;
   uc[pos].grpcou=0;
   uc[pos].grpsize = 0;
   print_grpsize(pos);
}


setup_ucfiles(pos)
int pos;
{
   if (strcmp(uc[pos].path, ".")==0)
   {
      getcwd(uc[pos].path, 198);
      shortpath(pos);
   }
   find_files(pos);
}


build_modestr(str, mode)
char *str;
int mode;
{
   int n;
   strcpy(str, "---------");
   for (n=0 ; n<9 ; n++)
   {
      if (fmode_values[n] & mode)
	 str[n]=fmode_valnames[n];
   }
}


read_modestr(str)
char *str;
{
   int n, mode=0;
   for (n=0 ; n<9 ; n++)
      if (str[n]!='-')
	 mode = mode | fmode_values[n];
   return(mode);
}


/************************************************************
   USER KEYS
************************************************************/

test_userkeys(pos)
int pos;
{
   int c, key = -1;
   char *cmd;
   int do_wait;

   prompt("[userkey]", "");
   c = ti_getch();
   if (c >= 'A'  &&  c < 'A'+UKNUM)
      key = c - 'A';
   if (c >= 'a'  &&  c < 'a'+UKNUM)
      key = c - 'a';
   if (c >= 1  &&  c < 1+UKNUM)
      key = c - 1;

   if (key != -1)
   {
      if (userkeys[key] != NULL)
      {
         cmd = assoc_macro(pos, userkeys[key], 
			   GETINFO(pos).name, &do_wait);
         do_fileexec(pos, cmd, do_wait);
      }
   }
}


/************************************************************
   FUNCTION KEYS
************************************************************/

do_functionkeys(pos)
int pos;
{
   int c;
   int ok=1;

   c = ti_getch();
   switch(c)
   {
      case '0' : ok = 0;
		 break;
      case '1' : do_chmod(pos);
		 break;
      case '2' : do_chown(pos);
		 break;
      case '3' : do_view(pos);
		 break;
      case '4' : do_edit(pos);
		 break;
      case '5' : do_copy(pos);
		 break;
      case '6' : do_move(pos);
		 break;
      case '7' : do_mkdir(pos);
		 break;
      case '8' : do_delete(pos);
		 break;
      case '9' : do_tohomedir(pos);
		 break;
      case 'l' :
      case 'L' : do_change_uc(0);
		 cursor(pos);
		 break;
      case 'r' :
      case 'R' : do_change_uc(1);
		 cursor(pos);
		 break;
      case 'h' :
      case 'H' : do_help(pos);
		 break;
      case 'c' :
      case 'C' : do_filecommand(pos);
		 cursor(pos);
		 break;
      case 'o' :
      case 'O' : do_options(pos);
		 cursor(pos);
		 break;
      case '+' : mark_group(pos);
		 break;
      case '-' : unmark_group(pos);
		 break;
   }
   return(ok);
}


do_chmod(pos)
int pos;
{
   char modestr[10];
   int mode, n;
   FILEINFO info;

   info = uc[pos].files[uc[pos].offset+uc[pos].pos];
   if (uc[pos].grpcou > 0)
      info.mode = 0777;
   build_modestr(modestr, info.mode);
   if (input_modestr(pos, info.name, modestr) == 0)
      return(0);

   mode = read_modestr(modestr);
   for (n=0 ; n<uc[pos].filenum ; n++)
   {
      if (uc[pos].files[n].mark)
      {
         if ( chmod(uc[pos].files[n].name, mode) != 0 )
            print_error();
         else
            uc[pos].files[n].mode = mode;
      }
   }

   if (uc[pos].grpcou == 0)
      GETINFO(pos).mark = 0;
   else
      print_files(pos);
   print_current(pos, uc[pos].pos, 1);

   return(1);
}


input_modestr(pos, name, str)
int pos;
char *name, *str;
{
   int start;
   int x=0, c;
   int ok=1, more=1;

   move(22,0);
   clrtoeol();
   if (uc[pos].grpcou == 0)
   {
      GETINFO(pos).mark = 1;
      printw("Change mode for %s : USR %c%c%c   GRP %c%c%c   ALL %c%c%c", 
	  name, str[0], str[1], str[2], str[3], str[4], 
	  str[5], str[6], str[7], str[8]);
      start = 23 + strlen(name);
   }
   else
   {
      printw("Change mode for %3d files : USR %c%c%c   GRP %c%c%c   ALL %c%c%c", 
	  uc[pos].grpcou, str[0], str[1], str[2], str[3], str[4], 
	  str[5], str[6], str[7], str[8]);
      start = 32;
   }

   do
   {
      move(22,start+x%3+(x/3)*10);
      refresh();
      c = ti_getch();
      if (c == terminal.enter)
         more=0;
      if (c == terminal.right)
	 if (x<8)
            x++;
      if (c == terminal.left)
         if (x>0)
	    x--;
      if (c=='r' || c=='w' || c=='x' || c==terminal.up)
      {
         str[x] = fmode_valnames[x];
         move(22,start+x%3+(x/3)*10);
         printw("%c", str[x]);
         if (x<8)
           x++;
      }
      if (c==' ' || c==terminal.down)
      {
         str[x] = '-';
         move(22,start+x%3+(x/3)*10);
         printw("%c", str[x]);
         if (x<8)
            x++;
      }
      if (c == 't')
      {
         if (str[x]=='-')
	    str[x] = fmode_valnames[x];
         else
	    str[x] = '-';
         move(22,start+x%3+(x/3)*10);
         printw("%c", str[x]);
	 if (x<8)
	    x++;
      }
      if (c==key_cc  ||  c==key_cd  ||  c=='q'  ||  c=='Q'  ||  c==27)
      {
         more = 0;
         ok = 0;
         if (uc[pos].grpcou == 0)
            GETINFO(pos).mark = 1;
      }
   }
   while (more);

   return(ok);
}


do_chown(pos)
int pos;
{
   char o[100], *name;
   int owner, group;
   struct group *grpe;
   struct passwd *pwde;
   int xpos=32, n;
   RETONTREE(pos);

   move(22,0);
   clrtoeol();
   if (uc[pos].grpcou == 0)
   {
      name = strdup(GETINFO(pos).name);
      xpos = 23+strlen(GETINFO(pos).name);
      if (xpos > 63)
      {
         name[40] = 0;
         xpos = 63;
      }
      GETINFO(pos).mark = 1;
      printw("Enter new owner for %s : ", name);
      free(name);
   }
   else
      printw("Enter new owner for %3d files : ", uc[pos].grpcou);   
   refresh();

   o[0] = 0;
   if (inp_input(xpos, 22, o, 10,30,0)==INPESC)
   {
      if (uc[pos].grpcou == 0)
         GETINFO(pos).mark = 0;
      return(0);
   }

   if (find_og(o, &owner, &group)==0)
   {
      move(22,0);
      clrtoeol();
      printw("ERROR : Owner not found.");
      refresh();
      ti_getch();
      if (uc[pos].grpcou == 0)
         GETINFO(pos).mark = 0;
      return(0);
   }
   group = GETINFO(pos).gid;
   
   for (n=0 ; n<uc[pos].filenum ; n++)
   {
      if (uc[pos].files[n].mark)
      {
         if ( chown(uc[pos].files[n].name, owner, group) != 0)
            print_error();
         else
         {
            uc[pos].files[n].uid = owner;
            if ( (pwde=getpwuid(owner)) != NULL )
               strcpy(uc[pos].files[n].usrname, pwde->pw_name);
            else
               strcpy(uc[pos].files[n].usrname, "----");

            if ( (grpe=getgrgid(group)) != NULL )
               strcpy(uc[pos].files[n].grpname, grpe->gr_name);
            else
               strcpy(uc[pos].files[n].grpname, "----");
         }
      }
   }

   if (uc[pos].grpcou == 0)
      GETINFO(pos).mark = 0;
   else
      print_files(pos);
   print_current(pos, uc[pos].pos, 1);
   
   return(1);
}


find_og(name, owner, group)
char *name;
int *owner, *group;
{
   struct passwd *pw;
   pw = getpwnam(name);
   if (pw==(struct passwd *)NULL)
   {
      return(0);
   }
   *owner = pw->pw_uid;
   *group = pw->pw_gid;
   return(1);
}



/************************************************************
   SEARCH FOR FILE
************************************************************/

search_file(pos, ypos)
int pos, ypos;
{
   static char buf[35] = {""};
   char *p=NULL;
   int y = ypos;
   int n, len, cmpval=1;

   prompt("Search: ", "");
   if (inp_input(8, 22, buf, 30,30,INPKEEP) == INPENTER)
   {
      if (uc[pos].sorting != SORT_NAME)
         qsort((void *)uc[pos].files, uc[pos].filenum, sizeof(FILEINFO),
	        sort_byname );

      len = strlen(buf);
      for (n=uc[pos].pos+uc[pos].offset+1 ; n<uc[pos].filenum  &&  cmpval!=0 ; n++)
      {
	 if ((cmpval=strncmp(uc[pos].files[n].name, buf, len)) <= 0)
            p = uc[pos].files[n].name;
      }

      if (uc[pos].sorting != SORT_NAME)
	 sort_files(pos);

      for (n=0 ; n<uc[pos].filenum ; n++)
	 if (uc[pos].files[n].name == p)
	    ypos = n;

      if (ypos>=uc[pos].offset  &&  ypos < uc[pos].offset+uc[pos].size)
	 ypos = ypos - uc[pos].offset;
      else
      {
	 uc[pos].offset = uc[pos].oldoffset = ypos;
	 uc[pos].pos = uc[pos].oldpos = 0;
	 ypos = 0;
         print_files(pos);
         print_dir(pos);
	 cursor(pos);
      }
   }

   return(ypos);
}

/************************************************************
   MKDIR
************************************************************/

do_mkdir(pos)
int pos;
{
   char dir[UCNAMESIZE+1];
   int dirmode;

   dirmode = S_IEXEC | S_IREAD | S_IWRITE | S_IRGRP | S_IROTH | S_IXGRP | S_IXOTH;
   prompt("Create directory :", "");
   dir[0]=0;
   if (inp_input(19,22,dir,61,UCNAMESIZE,0)==INPENTER)
   {
      if (chdir(uc[pos].path) != 0)
         print_error();
      if (mkdir(dir,uc[pos].dirmode)!=0)
	 print_error();

      if (uc[pos].showmode == SHOWTREE)
         mkt_update();
      else
         update_mkdir(pos, dir);
   }
}


update_mkdir(pos, dir)
int pos;
char *dir;
{
   int fpos; 

   find_files(pos);
   fpos = filename2pos(dir, pos);
   if (fpos > uc[pos].size)
      uc[pos].offset = fpos;
   else
      uc[pos].offset = 0;
   uc[pos].pos = fpos - uc[pos].offset;
   uc[pos].oldpos = -1;
   uc[pos].oldoffset = uc[pos].offset;

   update_screen(pos);
   if (strcmp(uc[pos].path, uc[N(pos)].path) == 0)
      update_screen(N(pos));

   cursor(pos);
}


/************************************************************
   GO TO HOME OR OTHER DIR.
************************************************************/

do_tohomedir(pos)
int pos;
{
   char *hdir;

   hdir = (char *)getenv("HOME");
   chdir(hdir);
   strcpy(uc[pos].path, hdir);
   shortpath(pos);

   uc[pos].pos = 0;
   uc[pos].oldpos = -1;
   uc[pos].offset = 0;
   uc[pos].oldoffset = 0;
   update_screen(pos);
   cursor(pos);
}


/************************************************************
   VIEW / EDIT
************************************************************/

do_view(pos)
int pos;
{
   char tmp[200];
   RETONTREE(pos);

   sprintf(tmp, "%s \"%s\"", prefvals[PREFVIEWER], GETINFO(pos).name);
   do_fileexec(pos, tmp, 0);

   return(1);
}


do_edit(pos)
int pos;
{
   char tmp[200];
   RETONTREE(pos);

   sprintf(tmp, "%s \"%s\"", prefvals[PREFEDITOR], GETINFO(pos).name);
   do_fileexec(pos, tmp, 0);

   return(1);
}


/************************************************************
   SWAP PANELS
************************************************************/

swappanels(pos)
int pos;
{
   FILESPEC swapspec;

   swapspec = uc[0];
   uc[0] = uc[1];
   uc[1] = swapspec;

   update_bscreens();
   cursor(N(pos));
}

/************************************************************
   MARK / UNMARK
************************************************************/

mark_group(pos)
int pos;
{
   int n;
   RETONTREE(pos);

   prompt("Select files : ", "");

   echo();
   if (inp_input(15+strlen(ucgrpmask),22, ucgrpmask, 20, 20,INPKEEP) == INPESC)
      return(0);

   noecho();
   for (n=0 ; n<uc[pos].filenum ; n++)
   {
      if (fname_match(uc[pos].files[n].name, ucgrpmask)
	  &&  (uc[pos].files[n].type==FTYPE_NORM 
	       || uc[pos].files[n].type==FTYPE_LINK))
      {
	 if (!uc[pos].files[n].mark)
	 {
   	    uc[pos].files[n].mark = 1;
	    uc[pos].grpcou++;
	    uc[pos].grpsize += uc[pos].files[n].size;
	 }
      }
   }
   print_files(pos);
   print_grpsize(pos);
   cursor(pos);

   return(1);
}


unmark_group(pos)
int pos;
{
   int n;
   RETONTREE(pos);

   prompt("Unselect files : ", "");

   echo();
   if (inp_input(17+strlen(ucgrpmask), 22, ucgrpmask, 20, 20,INPKEEP) == INPESC)
      return(0);

   noecho();
   for (n=0 ; n<uc[pos].filenum ; n++)
   {
      if (fname_match(uc[pos].files[n].name, ucgrpmask))
      {
	 if (uc[pos].files[n].mark)
	 {
	    uc[pos].files[n].mark = 0;
	    uc[pos].grpcou--;
	    uc[pos].grpsize -= uc[pos].files[n].size;
	 }
      }
   }
   print_files(pos);
   print_grpsize(pos);
   cursor(pos);

   return(1);
}


mark_one(pos)
int pos;
{
   RETONTREE(pos);

   if (GETINFO(pos).type==FTYPE_NORM || GETINFO(pos).type==FTYPE_LINK)
   {
      if ( (GETINFO(pos).mark=!GETINFO(pos).mark) == 1 )
      {
         uc[pos].grpcou++;
	 uc[pos].grpsize += GETINFO(pos).size;
      }
      else
      {
         uc[pos].grpcou--;
	 uc[pos].grpsize -= GETINFO(pos).size;
      }
      print_one_file(uc[pos].pos+3 ,pos, GETINFO(pos), 0);
   }
   if (uc[pos].pos+uc[pos].offset+1 < uc[pos].filenum)
      uc[pos].pos++;
   if (uc[pos].pos >= uc[pos].size)
   {
      uc[pos].pos = uc[pos].size-1;
      uc[pos].offset++;
      print_files(pos);
   }
   print_grpsize(pos);
   cursor(pos);

   return(1);
}


/************************************************************
   SHOW FILE INFO
************************************************************/

show_fileinfo(pos)
int pos;
{
   FILEINFO i;
   char modestr[20], oname[50], gname[50], linkname[59];
   struct group *grpe;
   struct passwd *pwde;
   int ys=5, yp=9, xp=4, len, islink=0;

#ifndef NOREADLINK
   if ((len=readlink(GETINFO(pos).name,linkname,58)) >= 0)
   {
      islink = 1;
      ys = 6;
      linkname[len] = 0;
   }
#endif /* NOREADLINK */

   uc_clear(yp,xp,ys,72);
   show_box(xp,yp, xp+72,yp+ys-1);

   i = GETINFO(pos);

   mvprintw(yp+1,xp+3, "FILE: %-58.58s", i.name);

#ifndef NOREADLINK
   if (islink)
      mvprintw(yp+2,xp+3, " -> : %-58.58s", linkname);
   else
      yp--;
#endif /* NOREADLINK */

   mvprintw(yp+3,xp+3, "TYPE:");
   if (i.type == FTYPE_NORM)
      mvprintw(yp+3,xp+9, "Norm file");
   else
   if (i.type == FTYPE_DIR)
      mvprintw(yp+3,xp+9, "Directory");
   else
   if (i.type == FTYPE_DEV)
      mvprintw(yp+3,xp+9, "Device");
   else
   if (i.type == FTYPE_FIFO)
      mvprintw(yp+3,xp+9, "Fifo");
   else
      mvprintw(yp+3,xp+9, "Other");

   build_modestr(modestr, i.mode);
   mvprintw(yp+4,xp+3, "MODE: %s", modestr);

   mvprintw(yp+3,xp+37, "SIZE: %-9d", i.size);

   mvprintw(yp+4,xp+37, "HARDLINKS: %d", i.linknum);

   mvprintw(yp+3,xp+20, "DATE: %02d/%02d/%02d",
	     i.fdate.tm_mday, i.fdate.tm_mon+1, i.fdate.tm_year%100);
   mvprintw(yp+4,xp+20, "TIME: %02d:%02d", 
	     i.fdate.tm_hour, i.fdate.tm_min);

   if ( (pwde=getpwuid(i.uid)) != NULL )
      mvprintw(yp+3,xp+54, "OWNER: %-8.8s", pwde->pw_name);
   else
      mvprintw(yp+3,xp+54, "OWNER: -----");

   if ( (grpe=getgrgid(i.gid)) != NULL )
      mvprintw(yp+4,xp+54, "GROUP: %-8.8s", grpe->gr_name);
   else
      mvprintw(yp+4,xp+54, "GROUP: -----");

   refresh();
   ti_getch();

   update2_bscreens();
}


uc_clear(y,x,l,c)
int y,x;
int l,c;
{
   int n,m;

   for (n=0 ; n<l ; n++)
      for (m=0 ; m<c ; m++)
	 mvaddch(n+y,m+x,' ');
}


/************************************************************
   Change UC setup
************************************************************/

do_change_uc(pos)
int pos;
{
   int c;

   prompt("Change :  L=Listning  S=Sorting  R=Reread  H=Hidden-files",">");
   c = ti_getch();
   if (c=='s'  ||  c=='S')
      select_sort(pos);
   if (c=='l'  ||  c=='L')
      change_listning(pos);
   if (c=='r'  ||  c=='R')
   {
      update_screen(pos);
      if (strcmp(uc[0].path, uc[1].path)==0)
	 update_screen(N(pos));
   }
   if (c=='h'  ||  c=='H')
   {
      uc[pos].showhidden = !uc[pos].showhidden;
      update_screen(pos);
   }
}


change_listning(pos)
int pos;
{
   int c;

   prompt("Listning: N=Normal  W=Wide  T=Tree",">");
   c = tolower(ti_getch());
   if (strchr("nwt", c) == NULL)
      return(0);
   
   if (c=='w')
   {
      uc[0].showmode = SHOWWIDE;
      uc[0].size = 8;
      uc[1].showmode = SHOWWIDE;
      uc[1].size = 8;
   }
   else
   if (c=='n')
   {
      uc[0].showmode = SHOWNORM;
      uc[0].size = 16;
      uc[1].showmode = SHOWNORM;
      uc[1].size = 16;
   }
   else
   if (c=='t')
   {
      uc[pos].showmode = SHOWTREE;
      prompt("Reading directory:","");
      test_maketree(pos,0);
   }

   clear();
   uc[pos].pos = 0;
   uc[pos].offset = 0;
   uc[pos].oldpos = -1;
   if (c!='t')
   {
      uc[N(pos)].offset = 0;
      uc[N(pos)].oldpos = -1;
      uc[N(pos)].pos = 0;
   }
   update_bscreens();

   return(1);
}


change_dir(pos)
int pos;
{
   static char buf[200] = {""};
   char *p;
   int y=0;

   prompt("Enter dir",">");
   if (inp_input(11,22, buf, 50,150,INPKEEP) == INPENTER)
   {
      p = subst_tilde(buf);
      if (chdir(p) == 0)
      {
	 getcwd(uc[pos].path, 198);
         shortpath(pos);
	 uc[pos].offset = uc[pos].oldoffset = 0;
	 uc[pos].pos = uc[pos].oldpos = 0;
	 update_screen(pos);
	 y = uc[pos].pos;
      }
      else
      {
         if (errno >= sys_nerr)
            prompt("Invalid directory", "");
         else
            prompt("Invalid directory: ", sys_errlist[errno]);
	 ti_getch();
      }
   }

   return(y);
}


char *subst_tilde(src)
char *src;
{
   static char buf[300];
   struct passwd *pw;
   char *d, name[50];
   int l,n,more;

   d = buf;
   while (*src != 0)
   {
      if (*src == '~')		/* Tilde substitute */
      {
	 src++;
	 l=0;
	 while (isalnum(*src)  &&  l<49)
	    name[l++] = *(src++);
	 name[l] = 0;

	 if ((pw=getpwnam(name)) != NULL)
	 {
	    l=0;
	    while (pw->pw_dir[l] != 0)
	    {
	       *(d++) = pw->pw_dir[l++];
	    }
	 }
      }
      else
      if (*src == '-')		/* Dir alias substititue */
      {
	 src++;
	 l=0;
	 while (isalnum(*src)  &&  l<49)
	    name[l++] = *(src++);
	 name[l] = 0;

	 more = 1;
	 for (n=0 ; n<daliasnum && more ; n++)
	 {
	    if (strcmp(diralias[n].mask, name) == 0)
	    {
	       more = 0;
	       for (l=0 ; diralias[n].command[l]!=0 ; l++)
		  *(d++) = diralias[n].command[l];
	    }
	 }
      }
      else
      {
	 *d = *src;
	 d++;
	 src++;
      }
   }
   *d = 0;

   return(buf);
}

/************************************************************
   COMMAND STACK
************************************************************/

push_command(str)
char *str;
{
   strcpy(command_stack[cstack_front], str);
   cstack_front = (cstack_front+1) % HISTORY;
   if (cstack_front == cstack_back)
      cstack_back = (cstack_back+1) % HISTORY;
   cstack_cur = cstack_front;
}


do_history_back()
{
   if (cstack_front == cstack_back  ||  cstack_cur == cstack_back)
      return(0);

   cstack_cur--;
   if (cstack_cur < 0)
      cstack_cur = HISTORY - 1;
   inp_clear(lineinp);
   inp_addstr(lineinp, command_stack[cstack_cur]);

   return(1);
}


do_history_forw()
{
   if (cstack_front == cstack_back  ||  cstack_cur == cstack_front)
      return(0);

   cstack_cur = (cstack_cur+1) % HISTORY;
   if (cstack_cur == cstack_front)
      return(0);
   inp_clear(lineinp);
   inp_addstr(lineinp, command_stack[cstack_cur]);

   return(1);
}


/************************************************************
   COMPARE
************************************************************/

do_compare()
{
   int n,m,pos;

   for (pos=0 ; pos<2 ; pos++)
   {
      for (n=0 ; n<uc[pos].filenum ; n++)
      {
	 if (uc[pos].files[n].type == FTYPE_NORM)
	 {
   	    uc[pos].files[n].mark = 1;
	    uc[pos].grpcou++;
	    uc[pos].grpsize += uc[pos].files[n].size;

	    for (m=0 ; m<uc[N(pos)].filenum ; m++)
	    {
	       if (cmp_equal(uc[pos].files[n], uc[N(pos)].files[m]))
	       {
	          uc[pos].files[n].mark = 0;
	          uc[pos].grpcou--;
	          uc[pos].grpsize -= uc[pos].files[n].size;
	       }
	    }
	 }
      }
   }
   
   print_files(0);
   print_files(1);
   print_grpsize(0);
   print_grpsize(1);
}


cmp_equal(a,b)
FILEINFO a,b;
{
   if (strcmp(a.name, b.name) != 0)
      return(0);
   if (a.size != b.size)
      return(0);

   return(1);
}

/************************************************************
   RELOAD
************************************************************/

do_reload(pos)
int pos;
{
   char cmd[200];

   sprintf(cmd, "%s %s/.uccfg", prefvals[PREFEDITOR], getenv("HOME"));
   do_fileexec(pos, cmd, 0);

   if (uc_yes("Reload configuration file"))
   {
      config_close();
      config_setup();
      find_cmd_options();
   }
}


/************************************************************
   HELP
************************************************************/

do_help(pos)
int pos;
{
   HTEXT *ht;

   if ( (ht=ht_new(HELPFILE, 22,80,1,0)) == NULL )
   {
      show_def_help(pos);
   }
   else
   {
      move(23,0);
      clrtoeol();
      attron(A_REVERSE);
      mvprintw(23,0, " Q = Quit ");
      attroff(A_REVERSE);
      ht_run(ht, "uchelp");
      ht_del(ht);
   }

   clr_display();
   draw_screen(0);
   draw_screen(1);
   draw_buttons();
   update2_bscreens();
   cursor(pos);
}


show_def_help(pos)
int pos;
{
   clr_display();
   mvaddstr(2,38,"HELP");
   mvaddstr(3,0,"--------------------------------------------------------------------------------");
   mvaddstr(21,0,"--------------------------------------------------------------------------------");

   mvaddstr(4,2,"CTRL +  B : Backward.");
   mvaddstr(5,10,"C : Clear.");
   mvaddstr(6,10,"D : Display fileinfo.");
   mvaddstr(7,10,"E : History.");
   mvaddstr(8,10,"F : Forward.");
   mvaddstr(9,10,"G : Select file.");
   mvaddstr(10,10,"J : Get filename.");
   mvaddstr(11,10,"K : Get current directory.");

   mvaddstr(4,40, "L : Last file.");
   mvaddstr(5,40, "N : New directory.");
   mvaddstr(6,40, "O : Search for file in dir.");
   mvaddstr(7,40, "P : User keys");
   mvaddstr(8,40, "R : Refresh files.");
   mvaddstr(9,40, "T : Top of files.");
   mvaddstr(10,40, "U : Swap display.");
   mvaddstr(11,40, "X : Forward history.");

   mvaddstr(13,2,"ESC/F1 followed by :");
   mvaddstr(14,10,"1..0 : Bottom line menu.");
   mvaddstr(15,10,"L : Setup left filelist.");
   mvaddstr(16,10,"R : Setup right filelist.");
   mvaddstr(17,10,"C : File commands.");
   mvaddstr(18,10,"O : Options.");
   mvaddstr(19,10,"+ : Select group.");
   mvaddstr(20,10,"- : Unselect group");
   prompt("Press any key","");
   ti_getch();

   clr_display();
   draw_screen(0);
   draw_screen(1);
   update2_bscreens();
   cursor(pos);
}

/************************************************************
   OPTIONS
************************************************************/

do_options(pos)
int pos;
{
   int c;
   prompt("OPTIONS:  K-Keyboard ", ">");
   c = ti_getch();
   if (c=='k'  ||  c=='K')
      keyboard_setup();
}


keyboard_setup()
{
   int c, more=0;
   clr_display();

   mvaddstr(2,29,"CHANGE KEYBOARD LAYOUT");
   mvaddstr(4,3,"Is it okay to proceed (y/n) ? ");
   refresh();
   c = ti_getch();

   if (c=='y'  ||  c=='Y')
      more = 1;

   while (more)
   {
      mvaddstr(4,33, "yes");
      mvaddstr(6,3, "Now you have to select what keys to use where ...");
      move(7,6);
      clrtoeol();
      addstr("Press key UP: ");
      refresh();
      c = terminal.up = ti_getch();
      printw("%03d", c);

      move(8,6);
      clrtoeol();
      addstr("Press key DOWN: ");
      refresh();
      c = terminal.down = ti_getch();
      printw("%03d", c);

      move(9,6);
      clrtoeol();
      addstr("Press key LEFT: ");
      refresh();
      c = terminal.left = ti_getch();
      printw("%03d", c);

      move(10,6);
      clrtoeol();
      addstr("Press key RIGHT: ");
      refresh();
      c = terminal.right = ti_getch();
      printw("%03d", c);

      move(11,6);
      clrtoeol();
      addstr("Press key ENTER: ");
      refresh();
      c = terminal.enter = ti_getch();
      printw("%03d", c);

      move(12,6);
      clrtoeol();
      addstr("Press key PAGEUP: ");
      refresh();
      c = terminal.pageup = ti_getch();
      printw("%03d", c);

      move(13,6);
      clrtoeol();
      addstr("Press key PAGEDOWN: ");
      refresh();
      c = terminal.pagedown = ti_getch();
      printw("%03d", c);

      move(14,6);
      clrtoeol();
      addstr("Press key BACKSPACE: ");
      refresh();
      c = terminal.backspace = ti_getch();
      printw("%03d", c);

      move(16,6);
      clrtoeol();
      addstr("Is this ok (y/n) ? ");
      refresh();
      c = ti_getch();

      more = 1;
      if (c=='y'  ||  c=='Y')
         more = 0;
   }

   clr_display();
   draw_screen(0);
   draw_screen(1);
   update2_bscreens();
}


#ifdef OPT_PREF
preferences_setup()
{
   int c;
   char buf[80];

   clr_display();
   mvaddstr(2,34,"PREFERENCES");
   mvaddstr(4,3, "Enter your prefered settings:");

   strcpy(buf, prefvals[PREFEDITOR]);
   mvaddstr(6,6, "Editor: ");
   if (inp_input(14,6, buf, 50,50,INPKEEP)==INPENTER)
   {
      free(prefvals[PREFEDITOR]);
      prefvals[PREFEDITOR] = strdup(buf);
   }

   strcpy(buf, prefvals[PREFVIEWER]);
   mvaddstr(7,6, "Viewer: ");
   if (inp_input(14,7, buf, 50,50,INPKEEP)==INPENTER)
   {
      free(prefvals[PREFVIEWER]);
      prefvals[PREFVIEWER] = strdup(buf);
   }

   mvaddstr(9,3, "Ok - press any key ... ");
   refresh();
   ti_getch();

   clr_display();
   draw_screen(0);
   draw_screen(1);
   update2_bscreens();
}
#endif


/************************************************************
   COMMANDS
************************************************************/

do_filecommand(pos)
int pos;
{
   int c;
   prompt("Select :  F=Find   C=Compare dir.  E=Edit&load cfgfile", ">");
   c = ti_getch();
   if (c=='f'  ||  c=='F')
      do_findfile(pos);
   if (c=='c'  ||  c=='C')
      do_compare();
   if (c=='e'  ||  c=='E')
      do_reload(pos);
}


do_enter(pos,y)
int pos,y;
{
   FILEINFO info;
   int fpos;
   int updir=0;
   char dirname[300];
   int ok=0;

   info = GETINFO(pos);
   if (lineinp->len > 0)  /* Er der skrevet noget paa kommandolinien */
   {
      do_fileexec(pos, input_str, 1);
      push_command(input_str);
      inp_clear(lineinp);
      y = uc[pos].pos;
   }
   else
   if (info.type == FTYPE_DIR)
   {
     if (uc[pos].showmode != SHOWTREE)
         ok = chdir(uc[pos].path);
      else
         ok = chdir(info.tree.path);

      if (ok != 0)
      {
         print_error();
         return(y);
      }

      if (uc[pos].showmode==SHOWTREE  &&  uc[N(pos)].showmode!=SHOWTREE)
      {
         strcpy(uc[N(pos)].path, info.tree.path);
         shortpath(N(pos));
         update_screen(N(pos));
         return(y);
      }

      shortpath(pos);
      path2dirname(uc[pos].path, dirname);
      if (strcmp(info.name, "..") == 0)
	 updir = 1;

      if (chdir(info.name) != 0)
         print_error();
      else
      {
         getcwd(uc[pos].path, 198);
         shortpath(pos);
         find_files(pos);
	 if (updir)
	 {
            fpos = filename2pos(dirname, pos);
            if (fpos > uc[pos].size)
	       uc[pos].offset = fpos;
            else
	       uc[pos].offset = 0;
            uc[pos].pos = fpos - uc[pos].offset;
            uc[pos].oldpos = -1;
            uc[pos].oldoffset = uc[pos].offset;
	 }
	 else
	 {
	    uc[pos].pos = 0;
	    uc[pos].oldpos = -1;
	    uc[pos].offset = uc[pos].oldoffset = 0;
	 }
         print_files(pos);
         y = uc[pos].pos;
         print_dir(pos);
      }
   }
   else
   if (info.mode & FMODE_EXEC)
   {
      sprintf(dirname, "./%s", GETINFO(pos).name);
      do_fileexec(pos, dirname, 1);
   }
   else
      assoc_test(pos, GETINFO(pos).name);

   return(y);
}


filename2pos(name, pos)
char *name;
int pos;
{
   int fpos=0;
   int n;
   for (n=0 ; n<uc[pos].filenum ; n++)
   {
      if (strcmp(name, uc[pos].files[n].name)==0)
	 fpos = n;
   }
   return(fpos);
}


path2dirname(path, name)
char *path, *name;
{
   int n;
   char *p;

   if ( (p=strrchr(path, '/')) == NULL )
      p = path;
   else
      p++;
   for (n=0 ; p[n]!=0 ; n++)
      name[n] = p[n];
   name[n] = 0;
}


do_fileexec(pos, comm, dowait)
int pos;
char *comm;
int dowait;
{
   int ppos, oldpos;
   int offset, oldoffset;

   clear();
   refresh();
   endwin();

   ch_cur_dir(pos);
   shortpath(pos);
   printf("%s>%s\n", get_cur_dir(pos), comm);
   uc_system(comm);

   raw();
   if (dowait)
   {
      printf("\nPress any key");
      fflush(stdout);
      ti_getch();
   }

   ppos = uc[pos].pos;
   oldpos = uc[pos].oldpos;
   offset = uc[pos].offset;
   oldoffset = uc[pos].oldoffset;

   update_bscreens();

   if (ppos+offset >= uc[pos].filenum)
   {
      offset = uc[pos].filenum/16;
      ppos = uc[pos].filenum - offset*16 - 1;
      oldpos = ppos;
      oldoffset = offset;
   }

   cursor(pos);
}


#define UCSYSTEM /**/
#ifdef UCSYSTEM
uc_system(cmd)
char *cmd;
{
   if (fork()==0)  /* child */
   {
      printf("\n");
      execl(prefvals[PREFSHELL], prefvals[PREFSHELL], prefshellopt, cmd, NULL);

      if (errno >= sys_nerr)
         printf("Unknown error : code %d\n", errno);
      else
         printf("ERROR : %s.\n", sys_errlist[errno]);
      printf("IN: %s %s %s\n", prefvals[PREFSHELL], prefshellopt, cmd);
      exit(0);
   }
   else  /* Parent */
      wait(NULL);   /* Wait for child */
}
#endif

